#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

#define BLOCK_SIZE 16

// Efficient CUDA kernel that performs GEMM using shared memory tiling with transposed weight load
// and fuses bias addition, scaling, and LeakyReLU activation.
__global__ void gemm_tiled_shared_kernel(
    const float* __restrict__ x,
    const float* __restrict__ weight,
    const float* __restrict__ bias,
    float* __restrict__ output,
    const int batch_size,
    const int in_features,
    const int out_features,
    const float multiplier,
    const float negative_slope
) {
    // Calculate global row and column indices
    int row = blockIdx.x * BLOCK_SIZE + threadIdx.x;
    int col = blockIdx.y * BLOCK_SIZE + threadIdx.y;

    float sum = 0.0f;

    // Shared memory tiles for x and transposed weight
    __shared__ float s_x[BLOCK_SIZE][BLOCK_SIZE];
    __shared__ float s_w[BLOCK_SIZE][BLOCK_SIZE];

    // Number of tiles needed along the in_features dimension
    const int numTiles = (in_features + BLOCK_SIZE - 1) / BLOCK_SIZE;

    for (int t = 0; t < numTiles; t++) {
        int tiled_index = t * BLOCK_SIZE;

        // Load tile from x into shared memory
        if (row < batch_size && (tiled_index + threadIdx.y) < in_features)
            s_x[threadIdx.x][threadIdx.y] = x[row * in_features + tiled_index + threadIdx.y];
        else
            s_x[threadIdx.x][threadIdx.y] = 0.0f;

        // Load tile from weight and transpose it for better coalescing
        // weight is stored in row-major with each row corresponding to an output feature
        if (col < out_features && (tiled_index + threadIdx.x) < in_features)
            s_w[threadIdx.y][threadIdx.x] = weight[col * in_features + tiled_index + threadIdx.x];
        else
            s_w[threadIdx.y][threadIdx.x] = 0.0f;

        __syncthreads();

        // Compute partial product for this tile
        #pragma unroll
        for (int k = 0; k < BLOCK_SIZE; k++) {
            sum += s_x[threadIdx.x][k] * s_w[threadIdx.y][k];
        }

        __syncthreads();
    }

    // Write the result back to global memory with bias addition, scaling and LeakyReLU
    if (row < batch_size && col < out_features) {
        sum = (sum + bias[col]) * multiplier;
        output[row * out_features + col] = (sum > 0.0f) ? sum : sum * negative_slope;
    }
}

// Host function to launch the kernel
torch::Tensor gemm_tiled_shared_forward(
    torch::Tensor x,
    float multiplier,
    float negative_slope,
    torch::Tensor weight,
    torch::Tensor bias
) {
    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(weight.is_cuda(), "weight must be a CUDA tensor");
    TORCH_CHECK(bias.is_cuda(), "bias must be a CUDA tensor");

    const int batch_size = x.size(0);
    const int in_features = x.size(1);
    const int out_features = weight.size(0);

    TORCH_CHECK(weight.size(1) == in_features, "Weight in_features must match x in_features");
    TORCH_CHECK(bias.size(0) == out_features, "Bias size must match weight out_features");

    auto output = torch::zeros({batch_size, out_features}, x.options());

    dim3 block(BLOCK_SIZE, BLOCK_SIZE);
    dim3 grid((batch_size + BLOCK_SIZE - 1) / BLOCK_SIZE,
              (out_features + BLOCK_SIZE - 1) / BLOCK_SIZE);

    gemm_tiled_shared_kernel<<<grid, block>>>(
        x.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.data_ptr<float>(),
        output.data_ptr<float>(),
        batch_size,
        in_features,
        out_features,
        multiplier,
        negative_slope
    );

    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &gemm_tiled_shared_forward, "Efficient GEMM with shared memory tiling, bias, scaling and LeakyReLU activation");
}
